@@ -15,6 +15,7 @@ from . import contrib
 from . import db
 from . import cite
 from . import errors
+from . import sync
 from .contrib import Project
 from .contrib import get_project
 from .contrib import init_project
@@ -33,9 +34,9 @@ from .contrib import MasterCrawler
 from .contrib import SignacProjectCrawler
 from .db import get_database
 
-__version__ = '0.8.7'
+__version__ = '0.9.0'
 
-__all__ = ['__version__', 'contrib', 'db', 'errors',
+__all__ = ['__version__', 'contrib', 'db', 'errors', 'sync',
            'cite',
            'Project', 'get_project', 'init_project',
            'get_database', 'fetch', 'fetch_one',
